/**************************************************************************
  BDD demonstration of counting.

**************************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include "bdd.h"

/* returns bdd for k out of n variabls with bdd_Var_ptr pointing to first
 * of the n variable.  The n bdd variables must be contiguous.
 */
bdd bdd_count(int k, int n, bdd *bdd_var)
{
  int i, j;
  bdd d[20];
  bdd carry, oldcarry;
  bdd answer = bddtrue;

  for (j = 0; j < 20; j++)
    d[j] = bddfalse;
  for (i = 0; i < n; i++) {
    carry = d[0] & bdd_var[i];  /* Carry if both bits set to 1. */
    d[0] = d[0] ^ bdd_var[i];   /* xor:  adding two digits in same column */

    for (j = 1; j < 20; j++) {
      oldcarry = carry;
      carry = d[j] & oldcarry; /* Carry if d1 is 1, and prev. had carry. */
      d[j] = d[j] ^ oldcarry; /* xor:  adding d2 and carry in same column */
    }
  }
  for (j = 0; j < 20; j++) {
    answer &= (k % 2 ? d[j] : !d[j]); /* check column of k and set d1 to same */
    k = k / 2; /* shift k to the right as binary number */
    n = n / 2; /* Stop counting after log_2 n iterations */
    if (n == 0)
      break;
  }
  if (j >= 20) {
    fprintf(stderr, "Maximum value of k is 2^20 = 1 million\n");
    abort();
  }
  return answer;
}

bdd bdd_satrandom(bdd bdd_arg, int var_num)
{
  int i, rand_int;
  bdd bdd_tmp = bdd_arg;

  srandom(time(NULL));
  for (i = 0; i < var_num; i++) {
    double count_true, count_false;
    count_true = bdd_satcount( bdd_tmp & bdd_ithvar(i) );
    count_false = bdd_satcount( bdd_tmp & !bdd_ithvar(i) );
    rand_int = random() & (1024 - 1);
    if (rand_int * (count_true + count_false) < 1024 * count_true)
      bdd_tmp = bdd_tmp & bdd_ithvar(i);
    else
      bdd_tmp = bdd_tmp & !bdd_ithvar(i);
  }
  return bdd_tmp;
}

int main(int argc, char **argv)
{
   using namespace std ;
   int k, n, i;
   
   if (argc != 3)
   {
      fprintf(stderr, "USAGE:  count k n\n");
      fprintf(stderr, "  Produce bdd with k out of n variables set true\n");
      return 1;
   }

   k = atoi(argv[1]);
   n = atoi(argv[2]);
   fprintf(stderr, "\nProducing bdd with %d out of %d variables set true\n",
	   k, n);

   /* Initialize with 100000 nodes, 10000 cache entries and n+2*k variables */
   bdd_init(n*256, 1000000);
   bdd_setvarnum(n);

   bdd x[10000]; /* We'll only use the first n in this array. */
   bdd answer;

   for (i = 0; i < n; i++)
     x[i] = bdd_ithvar(i);

   answer = bdd_count(k, n, &(x[0]));

      /* Print the results */
   cout << "There are " << bdd_satcount(answer) << " solutions.\n";
   cout << "One is:\n";
   bdd solution = bdd_satone(answer);
   cout << bddset << solution << endl;
   cout << "A random solution is:\n";
   solution = bdd_satrandom(answer, n);
   cout << bddset << solution << endl;
   cout << "This answer uses " << bdd_nodecount(answer) << " nodes.\n";

   bdd_done();
   
   return 0;
}
